# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"timer": "00_utils.ipynb",
         "filesize": "00_utils.ipynb",
         "save": "00_utils.ipynb",
         "load": "00_utils.ipynb",
         "push": "00_utils.ipynb",
         "refresh": "00_utils.ipynb",
         "backup": "00_utils.ipynb",
         "figParams": "00_utils.ipynb",
         "forceAspect": "00_utils.ipynb",
         "matshow": "00_utils.ipynb",
         "data": "01_conversion.ipynb",
         "trycast": "01_conversion.ipynb",
         "nonitr2itr": "01_conversion.ipynb",
         "itr2nonitr": "01_conversion.ipynb",
         "itr2itr": "01_conversion.ipynb",
         "Caster": "01_conversion.ipynb",
         "isiter": "01_conversion.ipynb",
         "cast": "01_conversion.ipynb",
         "typestr": "01_conversion.ipynb",
         "pad": "01_conversion.ipynb",
         "fill": "01_conversion.ipynb",
         "nbits": "01_conversion.ipynb",
         "num2ar": "01_conversion.ipynb",
         "ar2num": "01_conversion.ipynb",
         "ar2hex": "01_conversion.ipynb",
         "hex2ar": "01_conversion.ipynb",
         "str2ar": "01_conversion.ipynb",
         "ar2str": "01_conversion.ipynb",
         "COPY": "01_conversion.ipynb",
         "NOT": "01_conversion.ipynb",
         "AND": "01_conversion.ipynb",
         "OR": "01_conversion.ipynb",
         "Exclusive_OR": "01_conversion.ipynb",
         "XOR": "01_conversion.ipynb",
         "ar2gr": "01_conversion.ipynb",
         "gr2ar": "01_conversion.ipynb",
         "num2gr": "01_conversion.ipynb",
         "convert": "01_conversion.ipynb",
         "rint": "01_conversion.ipynb",
         "RNG": "01_conversion.ipynb",
         "depth": "02_recursion.ipynb",
         "flatten": "02_recursion.ipynb",
         "unflatten": "02_recursion.ipynb",
         "get": "02_recursion.ipynb",
         "give": "02_recursion.ipynb",
         "sort": "02_recursion.ipynb",
         "pipe": "02_recursion.ipynb",
         "maps": "02_recursion.ipynb",
         "replace": "03_templates.ipynb",
         "txt2lst": "03_templates.ipynb",
         "lst2txt": "03_templates.ipynb",
         "ZIP": "03_templates.ipynb",
         "getEvals": "03_templates.ipynb",
         "filltxt": "03_templates.ipynb",
         "Template": "03_templates.ipynb"}

modules = ["utils.py",
           "conversion.py",
           "recursion.py",
           "templates.py"]

doc_url = "https://Noeloikeau.github.io/sidis/"

git_url = "https://github.com/Noeloikeau/sidis/tree/master/"

def custom_doc_links(name): return None
